/**
 * @file geis_backend.h
 * @brief internal GEIS back end base class public interface
 *
 * Copyright 2010 Canonical Ltd.
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your option) any
 * later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */
#ifndef GEIS_BACKEND_H_
#define GEIS_BACKEND_H_

#include "geis/geis.h"
#include "geis_backend_token.h"

/**
 * Provides a virtual "base class" for various GEIS back ends.
 *
 * The GEIS is an API that implements the facade pattern over some number of
 * actual "back end" implementations.  Most internal operations of the API are
 * performed through this "virtual base class" interface rather than through the
 * concrete back ends (isn't object-oriented programming great?).
 */
typedef struct GeisBackend *GeisBackend;

/**
 * Creates a back end by name.
 *
 * @param[in] name  Names a back end.
 */
GeisBackend geis_backend_by_name(Geis geis, GeisString name);

/**
 * Destroys the back end.
 *
 * @parameter[in] be  The back end.
 *
 * This function behaves like a virtual destructor and chains through to the
 * concrete destructor call.
 */
void geis_backend_delete(GeisBackend be);

/**
 * Gets the name of the back end.
 *
 * @parameter[in] be  The back end.
 *
 * This accessor is useful for diagnostics.
 */
GeisString geis_backend_name(GeisBackend be);

/**
 * Creates a new back end token.
 */
GeisBackendToken geis_backend_create_token(GeisBackend be,
                                           GeisBackendTokenInitState);

/**
 * Marks a gesture as accepted by the back end.
 *
 * @param[in] be          The GEIS back end.
 * @param[in] group       The gesture group containing the accepted gesture.
 * @param[in] gesture_id  Identifies the gesture.
 */
GeisStatus
geis_backend_gesture_accept(GeisBackend   be,
                            GeisGroup     group,
                            GeisGestureId gesture_id);

/**
 * Marks a gesture as rejected by the back end.
 *
 * @param[in] be          The GEIS back end.
 * @param[in] group       The gesture group containing the rejected gesture.
 * @param[in] gesture_id  Identifies the gesture.
 */
GeisStatus
geis_backend_gesture_reject(GeisBackend   be,
                            GeisGroup     group,
                            GeisGestureId gesture_id);

#endif /* GEIS_BACKEND_H_ */
